@@ -13,7 +13,8 @@
 #include "log.h"
 #include "pool-buffer.h"
 #include "wlr-layer-shell-unstable-v1-client-protocol.h"
-#include "xdg-output-unstable-v1-client-protocol.h"
+#include "viewporter-client-protocol.h"
+#include "single-pixel-buffer-v1-client-protocol.h"
 
 static uint32_t parse_color(const char *color) {
 	if (color[0] == '#') {
@@ -38,7 +39,8 @@ struct swaybg_state {
 	struct wl_compositor *compositor;
 	struct wl_shm *shm;
 	struct zwlr_layer_shell_v1 *layer_shell;
-	struct zxdg_output_manager_v1 *xdg_output_manager;
+	struct wp_viewporter *viewporter;
+	struct wp_single_pixel_buffer_manager_v1 *single_pixel_buffer_manager;
 	struct wl_list configs;  // struct swaybg_output_config::link
 	struct wl_list outputs;  // struct swaybg_output::link
 	struct wl_list images;   // struct swaybg_image::link
@@ -63,7 +65,6 @@ struct swaybg_output_config {
 struct swaybg_output {
 	uint32_t wl_name;
 	struct wl_output *wl_output;
-	struct zxdg_output_v1 *xdg_output;
 	char *name;
 	char *identifier;
 
@@ -117,6 +118,35 @@ static void render_frame(struct swaybg_output *output, cairo_surface_t *surface)
 		}
 		return;
 	}
+
+	if (output->config->mode == BACKGROUND_MODE_SOLID_COLOR &&
+			output->state->viewporter &&
+			output->state->single_pixel_buffer_manager) {
+		uint8_t r8 = (output->config->color >> 24) & 0xFF;
+		uint8_t g8 = (output->config->color >> 16) & 0xFF;
+		uint8_t b8 = (output->config->color >> 8) & 0xFF;
+		uint8_t a8 = (output->config->color >> 0) & 0xFF;
+		uint32_t f = 0xFFFFFFFF / 0xFF; // division result is an integer
+		uint32_t r32 = r8 * f;
+		uint32_t g32 = g8 * f;
+		uint32_t b32 = b8 * f;
+		uint32_t a32 = a8 * f;
+		struct wl_buffer *buffer = wp_single_pixel_buffer_manager_v1_create_u32_rgba_buffer(
+			output->state->single_pixel_buffer_manager, r32, g32, b32, a32);
+		wl_surface_attach(output->surface, buffer, 0, 0);
+		wl_surface_damage_buffer(output->surface, 0, 0, INT32_MAX, INT32_MAX);
+
+		struct wp_viewport *viewport = wp_viewporter_get_viewport(
+			output->state->viewporter, output->surface);
+		wp_viewport_set_destination(viewport, output->width, output->height);
+
+		wl_surface_commit(output->surface);
+
+		wp_viewport_destroy(viewport);
+		wl_buffer_destroy(buffer);
+		return;
+	}
+
 	struct pool_buffer buffer;
 	if (!create_buffer(&buffer, output->state->shm,
 			buffer_width, buffer_height, WL_SHM_FORMAT_ARGB8888)) {
@@ -184,7 +214,6 @@ static void destroy_swaybg_output(struct swaybg_output *output) {
 	if (output->surface != NULL) {
 		wl_surface_destroy(output->surface);
 	}
-	zxdg_output_v1_destroy(output->xdg_output);
 	wl_output_destroy(output->wl_output);
 	free(output->name);
 	free(output->identifier);
@@ -226,81 +255,6 @@ static void output_mode(void *data, struct wl_output *output, uint32_t flags,
 	// Who cares
 }
 
-static void output_done(void *data, struct wl_output *output) {
-	// Who cares
-}
-
-static void output_scale(void *data, struct wl_output *wl_output,
-		int32_t scale) {
-	struct swaybg_output *output = data;
-	output->scale = scale;
-	if (output->state->run_display && output->width > 0 && output->height > 0) {
-		output->dirty = true;
-	}
-}
-
-static const struct wl_output_listener output_listener = {
-	.geometry = output_geometry,
-	.mode = output_mode,
-	.done = output_done,
-	.scale = output_scale,
-};
-
-static void xdg_output_handle_logical_position(void *data,
-		struct zxdg_output_v1 *xdg_output, int32_t x, int32_t y) {
-	// Who cares
-}
-
-static void xdg_output_handle_logical_size(void *data,
-		struct zxdg_output_v1 *xdg_output, int32_t width, int32_t height) {
-	// Who cares
-}
-
-static void find_config(struct swaybg_output *output, const char *name) {
-	struct swaybg_output_config *config = NULL;
-	wl_list_for_each(config, &output->state->configs, link) {
-		if (strcmp(config->output, name) == 0) {
-			output->config = config;
-			return;
-		} else if (!output->config && strcmp(config->output, "*") == 0) {
-			output->config = config;
-		}
-	}
-}
-
-static void xdg_output_handle_name(void *data,
-		struct zxdg_output_v1 *xdg_output, const char *name) {
-	struct swaybg_output *output = data;
-	output->name = strdup(name);
-
-	// If description was sent first, the config may already be populated. If
-	// there is an identifier config set, keep it.
-	if (!output->config || strcmp(output->config->output, "*") == 0) {
-		find_config(output, name);
-	}
-}
-
-static void xdg_output_handle_description(void *data,
-		struct zxdg_output_v1 *xdg_output, const char *description) {
-	struct swaybg_output *output = data;
-
-	// wlroots currently sets the description to `make model serial (name)`
-	// If this changes in the future, this will need to be modified.
-	char *paren = strrchr(description, '(');
-	if (paren) {
-		size_t length = paren - description;
-		output->identifier = malloc(length);
-		if (!output->identifier) {
-			swaybg_log(LOG_ERROR, "Failed to allocate output identifier");
-			return;
-		}
-		strncpy(output->identifier, description, length);
-		output->identifier[length - 1] = '\0';
-
-		find_config(output, output->identifier);
-	}
-}
-
 static void create_layer_surface(struct swaybg_output *output) {
 	output->surface = wl_compositor_create_surface(output->state->compositor);
 	assert(output->surface);
@@ -329,8 +283,7 @@ static void create_layer_surface(struct swaybg_output *output) {
 	wl_surface_commit(output->surface);
 }
 
-static void xdg_output_handle_done(void *data,
-		struct zxdg_output_v1 *xdg_output) {
+static void output_done(void *data, struct wl_output *wl_output) {
 	struct swaybg_output *output = data;
 	if (!output->config) {
 		swaybg_log(LOG_DEBUG, "Could not find config for output %s (%s)",
@@ -343,12 +296,67 @@ static void xdg_output_handle_done(void *data,
 	}
 }
 
-static const struct zxdg_output_v1_listener xdg_output_listener = {
-	.logical_position = xdg_output_handle_logical_position,
-	.logical_size = xdg_output_handle_logical_size,
-	.name = xdg_output_handle_name,
-	.description = xdg_output_handle_description,
-	.done = xdg_output_handle_done,
+static void output_scale(void *data, struct wl_output *wl_output,
+		int32_t scale) {
+	struct swaybg_output *output = data;
+	output->scale = scale;
+	if (output->state->run_display && output->width > 0 && output->height > 0) {
+		output->dirty = true;
+	}
+}
+
+static void find_config(struct swaybg_output *output, const char *name) {
+	struct swaybg_output_config *config = NULL;
+	wl_list_for_each(config, &output->state->configs, link) {
+		if (strcmp(config->output, name) == 0) {
+			output->config = config;
+			return;
+		} else if (!output->config && strcmp(config->output, "*") == 0) {
+			output->config = config;
+		}
+	}
+}
+
+static void output_name(void *data, struct wl_output *wl_output,
+		const char *name) {
+	struct swaybg_output *output = data;
+	output->name = strdup(name);
+
+	// If description was sent first, the config may already be populated. If
+	// there is an identifier config set, keep it.
+	if (!output->config || strcmp(output->config->output, "*") == 0) {
+		find_config(output, name);
+	}
+}
+
+static void output_description(void *data, struct wl_output *wl_output,
+		const char *description) {
+	struct swaybg_output *output = data;
+
+	// wlroots currently sets the description to `make model serial (name)`
+	// If this changes in the future, this will need to be modified.
+	char *paren = strrchr(description, '(');
+	if (paren) {
+		size_t length = paren - description;
+		output->identifier = malloc(length);
+		if (!output->identifier) {
+			swaybg_log(LOG_ERROR, "Failed to allocate output identifier");
+			return;
+		}
+		strncpy(output->identifier, description, length);
+		output->identifier[length - 1] = '\0';
+
+		find_config(output, output->identifier);
+	}
+}
+
+static const struct wl_output_listener output_listener = {
+	.geometry = output_geometry,
+	.mode = output_mode,
+	.done = output_done,
+	.scale = output_scale,
+	.name = output_name,
+	.description = output_description,
 };
 
 static void handle_global(void *data, struct wl_registry *registry,
@@ -364,22 +372,19 @@ static void handle_global(void *data, struct wl_registry *registry,
 		output->state = state;
 		output->wl_name = name;
 		output->wl_output =
-			wl_registry_bind(registry, name, &wl_output_interface, 3);
+			wl_registry_bind(registry, name, &wl_output_interface, 4);
 		wl_output_add_listener(output->wl_output, &output_listener, output);
 		wl_list_insert(&state->outputs, &output->link);
-
-		if (state->run_display) {
-			output->xdg_output = zxdg_output_manager_v1_get_xdg_output(
-				state->xdg_output_manager, output->wl_output);
-			zxdg_output_v1_add_listener(output->xdg_output,
-				&xdg_output_listener, output);
-		}
 	} else if (strcmp(interface, zwlr_layer_shell_v1_interface.name) == 0) {
 		state->layer_shell =
 			wl_registry_bind(registry, name, &zwlr_layer_shell_v1_interface, 1);
-	} else if (strcmp(interface, zxdg_output_manager_v1_interface.name) == 0) {
-		state->xdg_output_manager = wl_registry_bind(registry, name,
-			&zxdg_output_manager_v1_interface, 2);
+	} else if (strcmp(interface, wp_viewporter_interface.name) == 0) {
+		state->viewporter = wl_registry_bind(registry, name,
+			&wp_viewporter_interface, 1);
+	} else if (strcmp(interface,
+			wp_single_pixel_buffer_manager_v1_interface.name) == 0) {
+		state->single_pixel_buffer_manager = wl_registry_bind(registry, name,
+			&wp_single_pixel_buffer_manager_v1_interface, 1);
 	}
 }
 
@@ -573,24 +578,20 @@ int main(int argc, char **argv) {
 
 	struct wl_registry *registry = wl_display_get_registry(state.display);
 	wl_registry_add_listener(registry, &registry_listener, &state);
-	wl_display_roundtrip(state.display);
-	if (state.compositor == NULL || state.shm == NULL ||
-			state.layer_shell == NULL || state.xdg_output_manager == NULL) {
-		swaybg_log(LOG_ERROR, "Missing a required Wayland interface");
+	if (wl_display_roundtrip(state.display) < 0) {
+		swaybg_log(LOG_ERROR, "wl_display_roundtrip failed");
 		return 1;
 	}
-
-	struct swaybg_output *output;
-	wl_list_for_each(output, &state.outputs, link) {
-		output->xdg_output = zxdg_output_manager_v1_get_xdg_output(
-			state.xdg_output_manager, output->wl_output);
-		zxdg_output_v1_add_listener(output->xdg_output,
-			&xdg_output_listener, output);
+	if (state.compositor == NULL || state.shm == NULL ||
+			state.layer_shell == NULL) {
+		swaybg_log(LOG_ERROR, "Missing a required Wayland interface");
+		return 1;
 	}
 
 	state.run_display = true;
 	while (wl_display_dispatch(state.display) != -1 && state.run_display) {
 		// Send acks, and determine which images need to be loaded
+		struct swaybg_output *output;
 		wl_list_for_each(output, &state.outputs, link) {
 			if (output->needs_ack) {
 				output->needs_ack = false;
@@ -641,7 +642,7 @@ int main(int argc, char **argv) {
 		}
 	}
 
-	struct swaybg_output *tmp_output;
+	struct swaybg_output *output, *tmp_output;
 	wl_list_for_each_safe(output, tmp_output, &state.outputs, link) {
 		destroy_swaybg_output(output);
 	}
